@@ -12,7 +12,7 @@ mkdir "$DIST"/.cargo
 cargo vendor-filterer --all-features --platform=x86_64-unknown-linux-gnu --platform=aarch64-unknown-linux-gnu > "$DIST"/.cargo/config.toml
 rm -f vendor/gettext-sys/gettext-*.tar.*
 # remove the tarball from checksums
-echo $(jq -c 'del(.files["gettext-0.22.tar.xz"])' vendor/gettext-sys/.cargo-checksum.json) > vendor/gettext-sys/.cargo-checksum.json
+echo $(jq -c 'del(.files["gettext-0.22.5.tar.xz"])' vendor/gettext-sys/.cargo-checksum.json) > vendor/gettext-sys/.cargo-checksum.json
 # Don't combine the previous and this line with a pipe because we can't catch
 # errors with "set -o pipefail"
 sed -i 's/^directory = ".*"/directory = "vendor"/g' "$DIST/.cargo/config.toml"
