@@ -5,16 +5,29 @@ use gettextrs::gettext;
 use gtk::CompositeTemplate;
 use gtk::{gio, glib};
 
+#[cfg(feature = "portal")]
+use anyhow::Context;
+#[cfg(feature = "portal")]
+use ashpd::desktop::camera;
+#[cfg(feature = "portal")]
+use std::os::unix::io::OwnedFd;
+
 use super::CameraControls;
+use crate::enums::ControlsLayout;
 use crate::{config, utils};
 
 mod imp {
     use std::cell::{Cell, OnceCell, RefCell};
 
+    use gtk::{CallbackAction, Shortcut, ShortcutController, ShortcutTrigger};
+
+    use crate::CaptureMode;
+
     use super::*;
 
-    #[derive(Debug, Default, CompositeTemplate)]
+    #[derive(Debug, Default, CompositeTemplate, glib::Properties)]
     #[template(resource = "/org/gnome/Snapshot/ui/camera.ui")]
+    #[properties(wrapper_type = super::Camera)]
     pub struct Camera {
         pub selection: gtk::SingleSelection,
         pub provider: OnceCell<aperture::DeviceProvider>,
@@ -25,6 +38,9 @@ mod imp {
         pub recording_duration: Cell<u32>,
         pub recording_source: RefCell<Option<glib::source::SourceId>>,
 
+        #[property(get, set = Self::set_capture_mode, explicit_notify, builder(Default::default()))]
+        capture_mode: Cell<crate::CaptureMode>,
+
         #[template_child]
         pub single_landscape_bp: TemplateChild<adw::Breakpoint>,
         #[template_child]
@@ -42,25 +58,20 @@ mod imp {
         #[template_child]
         pub flash_bin: TemplateChild<crate::FlashBin>,
         #[template_child]
-        pub stack: TemplateChild<gtk::Stack>,
+        pub qr_screen_bin: TemplateChild<crate::QrScreenBin>,
         #[template_child]
-        pub spinner: TemplateChild<gtk::Spinner>,
+        pub stack: TemplateChild<gtk::Stack>,
 
         #[template_child]
         pub guidelines: TemplateChild<crate::GuidelinesBin>,
 
         #[template_child]
-        pub camera_controls_vertical: TemplateChild<crate::CameraControls>,
-        #[template_child]
-        pub camera_controls_horizontal: TemplateChild<crate::CameraControls>,
+        pub camera_controls: TemplateChild<crate::CameraControls>,
 
         #[template_child]
-        pub sidebar_horizontal_end: TemplateChild<gtk::CenterBox>,
+        pub bottom_sheet: TemplateChild<adw::BottomSheet>,
         #[template_child]
-        pub sidebar_vertical_end: TemplateChild<gtk::CenterBox>,
-
-        #[template_child]
-        pub vertical_end_window_controls: TemplateChild<gtk::WindowControls>,
+        pub qr_bottom_sheet: TemplateChild<crate::QrBottomSheet>,
     }
 
     #[glib::object_subclass]
@@ -82,6 +93,24 @@ mod imp {
 
     #[gtk::template_callbacks]
     impl Camera {
+        fn set_capture_mode(&self, capture_mode: crate::CaptureMode) {
+            if capture_mode != self.capture_mode.replace(capture_mode) {
+                match capture_mode {
+                    CaptureMode::Picture => {
+                        self.obj().set_shutter_mode(crate::ShutterMode::Picture);
+                    }
+                    CaptureMode::Video => {
+                        self.obj().set_shutter_mode(crate::ShutterMode::Video);
+                    }
+                    CaptureMode::QrDetection => (),
+                };
+                self.obj()
+                    .set_detect_codes(matches!(capture_mode, CaptureMode::QrDetection));
+
+                self.obj().notify_capture_mode();
+            }
+        }
+
         pub fn settings(&self) -> &gio::Settings {
             self.settings
                 .get_or_init(|| gio::Settings::new(config::APP_ID))
@@ -101,6 +130,7 @@ mod imp {
         }
     }
 
+    #[glib::derived_properties]
     impl ObjectImpl for Camera {
         fn constructed(&self) {
             self.parent_constructed();
@@ -110,48 +140,120 @@ mod imp {
             let provider = aperture::DeviceProvider::instance();
             self.provider.set(provider.clone()).unwrap();
 
-            provider.connect_camera_added(glib::clone!(@weak obj => move |provider, _| {
-                obj.update_cameras_button(provider);
-            }));
-            provider.connect_camera_removed(glib::clone!(@weak obj => move |provider, _| {
-                obj.update_cameras_button(provider);
-            }));
+            let create_shortcut = |shortcut, value: CaptureMode| {
+                Shortcut::new(
+                    ShortcutTrigger::parse_string(shortcut),
+                    Some(CallbackAction::new(glib::clone!(
+                        #[weak]
+                        obj,
+                        #[upgrade_or]
+                        glib::Propagation::Proceed,
+                        move |_, _| {
+                            obj.set_capture_mode(value);
+                            glib::Propagation::Proceed
+                        }
+                    ))),
+                )
+            };
+
+            let controller = ShortcutController::new();
+            controller.set_scope(gtk::ShortcutScope::Managed);
+            controller.add_shortcut(create_shortcut("p", CaptureMode::Picture));
+            controller.add_shortcut(create_shortcut("r", CaptureMode::Video));
+
+            obj.add_controller(controller);
+
+            provider.connect_camera_added(glib::clone!(
+                #[weak]
+                obj,
+                move |provider, _| {
+                    obj.update_cameras_button(provider);
+                }
+            ));
+            provider.connect_camera_removed(glib::clone!(
+                #[weak]
+                obj,
+                move |provider, _| {
+                    obj.update_cameras_button(provider);
+                }
+            ));
             obj.update_cameras_button(provider);
 
-            self.viewfinder
-                .connect_state_notify(glib::clone!(@weak obj => move |_| {
+            self.viewfinder.connect_state_notify(glib::clone!(
+                #[weak]
+                obj,
+                move |_| {
                     obj.update_state();
-                }));
+                }
+            ));
+
+            self.viewfinder.connect_code_detected(glib::clone!(
+                #[weak]
+                obj,
+                move |_, code| {
+                    match std::str::from_utf8(&code) {
+                        Ok(code) => {
+                            log::debug!("Detected QR code: {code}");
+                            obj.imp().bottom_sheet.set_open(true);
+                            obj.imp().qr_bottom_sheet.set_contents(code);
+                        }
+                        Err(err) => {
+                            log::error!("Could not decode QR code into utf8: {err}");
+                        }
+                    }
+                }
+            ));
+
+            self.qr_screen_bin.set_viewfinder(self.viewfinder.clone());
+
             obj.update_state();
 
+            self.viewfinder.connect_is_recording_notify(glib::clone!(
+                #[weak]
+                obj,
+                move |viewfinder| {
+                    let window = viewfinder.root().and_downcast::<crate::Window>().unwrap();
+
+                    if viewfinder.is_recording() {
+                        obj.set_shutter_mode(crate::ShutterMode::Recording);
+                        window.inhibit("Recording Video");
+                        obj.show_recording_label();
+                    } else {
+                        obj.hide_recording_label();
+                        window.uninhibit();
+                        if matches!(obj.shutter_mode(), crate::ShutterMode::Recording) {
+                            obj.set_shutter_mode(crate::ShutterMode::Video);
+                        }
+                    }
+                }
+            ));
+
             self.selection.set_model(Some(provider));
-            self.selection.connect_selected_item_notify(
-                glib::clone!(@weak obj => move |selection| {
+            self.selection.connect_selected_item_notify(glib::clone!(
+                #[weak]
+                obj,
+                move |selection| {
                     if let Some(selected_item) = selection.selected_item() {
                         let camera = selected_item.downcast::<aperture::Camera>().ok();
 
-                        if matches!(obj.imp().viewfinder.state(), aperture::ViewfinderState::Ready | aperture::ViewfinderState::Error) {
+                        if matches!(
+                            obj.imp().viewfinder.state(),
+                            aperture::ViewfinderState::Ready | aperture::ViewfinderState::Error
+                        ) {
                             obj.set_camera_inner(camera);
                         }
                     }
-                }),
-            );
+                }
+            ));
 
-            self.camera_controls_horizontal
-                .set_selection(self.selection.clone());
-            self.camera_controls_vertical
-                .set_selection(self.selection.clone());
-
-            self.camera_controls_horizontal.connect_camera_switched(
-                glib::clone!(@weak self as obj => move |_: &CameraControls| {
+            self.camera_controls.set_selection(self.selection.clone());
+            self.camera_controls.connect_camera_switched(glib::clone!(
+                #[weak(rename_to = obj)]
+                self,
+                move |_: &CameraControls| {
                     obj.obj().camera_switched();
-                }),
-            );
-            self.camera_controls_vertical.connect_camera_switched(
-                glib::clone!(@weak self as obj => move |_: &CameraControls| {
-                    obj.obj().camera_switched();
-                }),
-            );
+                }
+            ));
 
             self.settings()
                 .bind(
@@ -161,27 +263,34 @@ mod imp {
                 )
                 .build();
 
-            // TODO remove if
-            // https://gitlab.gnome.org/GNOME/gtk/-/merge_requests/5960 ever
-            // lands.
-            obj.update_window_controls();
-            obj.settings().connect_gtk_decoration_layout_notify(
-                glib::clone!(@weak obj => move |_| {
-                    obj.update_window_controls();
-                }),
-            );
+            self.settings()
+                .bind(
+                    "enable-audio-recording",
+                    &*self.viewfinder,
+                    "disable-audio-recording",
+                )
+                .invert_boolean()
+                .build();
 
-            obj.connect_current_breakpoint_notify(glib::clone!(@weak self as obj => move |imp| {
-                if imp.current_breakpoint().is_none()
-                || imp
-                    .current_breakpoint()
-                    .is_some_and(|breakpoint| breakpoint.eq(&obj.dual_portrait_bp.get()))
-                {
-                    imp.add_css_class("portrait");
-                } else {
-                    imp.remove_css_class("portrait");
+            self.settings()
+                .bind("capture-mode", &*obj, "capture-mode")
+                .build();
+
+            obj.connect_current_breakpoint_notify(glib::clone!(
+                #[weak(rename_to = obj)]
+                self,
+                move |imp| {
+                    if imp.current_breakpoint().is_none()
+                        || imp
+                            .current_breakpoint()
+                            .is_some_and(|breakpoint| breakpoint.eq(&obj.dual_portrait_bp.get()))
+                    {
+                        imp.add_css_class("portrait");
+                    } else {
+                        imp.remove_css_class("portrait");
+                    }
                 }
-            }));
+            ));
         }
     }
 
@@ -205,22 +314,73 @@ impl Camera {
         Self::default()
     }
 
+    #[cfg(feature = "portal")]
+    fn on_portal_not_allowed(&self) {
+        // We don't start the device provider if we are not
+        // allowed to use cameras.
+        self.imp().permission_denied.set(true);
+        self.update_state();
+    }
+
     pub async fn start(&self) {
         let provider = self.imp().provider.get().unwrap();
 
-        glib::spawn_future_local(
-            glib::clone!(@weak self as obj, @strong provider => async move {
-                if let Err(err) = provider.start_with_default(glib::clone!(@weak obj => @default-return false, move |camera| {
-                    let stored_id = obj.imp().settings().string("last-camera-id");
-                    !stored_id.is_empty() && id_from_pw(camera) == stored_id
-                })) {
+        glib::spawn_future_local(glib::clone!(
+            #[weak(rename_to = obj)]
+            self,
+            #[strong]
+            provider,
+            async move {
+                #[cfg(feature = "portal")]
+                if let Err(err) = ashpd::register_host_app(config::APP_ID.try_into().unwrap()).await
+                {
+                    log::error!(
+                        "Failed to run org.freedesktop.host.portal.Registry.Register: {err}"
+                    );
+                }
+                #[cfg(feature = "portal")]
+                match stream().await {
+                    Ok(fd) => {
+                        if let Err(err) = provider.set_fd(fd) {
+                            log::error!("Could not use the camera portal: {err}");
+                        };
+                    }
+                    Err(err) => match err.downcast_ref::<ashpd::Error>() {
+                        Some(ashpd::Error::Portal(ashpd::PortalError::NotAllowed(err))) => {
+                            log::warn!("Permission to use the camera portal denied: {err:#?}");
+                            obj.on_portal_not_allowed();
+                            return;
+                        }
+                        Some(ashpd::Error::Zbus(ashpd::zbus::Error::MethodError(
+                            name,
+                            _,
+                            message,
+                        ))) if *name == "org.freedesktop.portal.Error.NotAllowed" => {
+                            log::warn!("Permission to use the camera portal denied: {message}");
+                            obj.on_portal_not_allowed();
+                            return;
+                        }
+                        _ => (),
+                    },
+                }
+
+                if let Err(err) = provider.start_with_default(glib::clone!(
+                    #[weak]
+                    obj,
+                    #[upgrade_or]
+                    false,
+                    move |camera| {
+                        let stored_id = obj.imp().settings().string("last-camera-id");
+                        !stored_id.is_empty() && id_from_pw(camera) == stored_id
+                    }
+                )) {
                     log::error!("Could not start the device provider: {err}");
                 } else {
                     log::debug!("Device provider started");
                     obj.update_cameras_button(&provider);
                 };
-            }),
-        );
+            }
+        ));
     }
 
     pub async fn start_recording(&self, format: crate::VideoFormat) -> anyhow::Result<()> {
@@ -228,7 +388,6 @@ impl Camera {
         let path = utils::videos_dir()?.join(filename);
 
         self.imp().viewfinder.start_recording(path)?;
-        self.show_recording_label();
 
         Ok(())
     }
@@ -241,7 +400,6 @@ impl Camera {
             if let Err(err) = imp.viewfinder.stop_recording() {
                 log::error!("Could not stop camera: {err}");
             }
-            self.hide_recording_label();
         }
     }
 
@@ -249,7 +407,7 @@ impl Camera {
         let imp = self.imp();
         let window = self.root().and_downcast::<crate::Window>().unwrap();
 
-        // We enable the shutter whenever picture-stored is emited.
+        // We enable the shutter whenever picture-stored is emitted.
         window.set_shutter_enabled(false);
 
         let filename = utils::picture_file_name(format);
@@ -288,6 +446,10 @@ impl Camera {
             imp.settings().set_string("last-camera-id", &id).unwrap();
         }
 
+        if imp.viewfinder.is_recording() {
+            self.stop_recording();
+        }
+
         imp.viewfinder.set_camera(camera);
     }
 
@@ -301,81 +463,72 @@ impl Camera {
         self.imp().players.replace(Some(player));
     }
 
-    fn active_controls(&self) -> &CameraControls {
-        let imp = self.imp();
-        let bp1 = &*imp.single_landscape_bp;
-        let bp2 = &*imp.dual_landscape_bp;
-        if self
-            .current_breakpoint()
-            .is_some_and(|bp| &bp == bp1 || &bp == bp2)
-        {
-            &imp.camera_controls_horizontal
-        } else {
-            &imp.camera_controls_vertical
-        }
-    }
-
     pub fn set_countdown(&self, countdown: u32) {
-        self.imp()
-            .camera_controls_horizontal
-            .set_countdown(countdown);
-        self.imp().camera_controls_vertical.set_countdown(countdown);
+        self.imp().camera_controls.set_countdown(countdown);
     }
 
     pub fn start_countdown(&self) {
-        self.imp().camera_controls_horizontal.start_countdown();
-        self.imp().camera_controls_vertical.start_countdown();
+        self.imp().camera_controls.start_countdown();
     }
 
     pub fn stop_countdown(&self) {
-        self.imp().camera_controls_horizontal.stop_countdown();
-        self.imp().camera_controls_vertical.stop_countdown();
+        self.imp().camera_controls.stop_countdown();
     }
 
     pub fn shutter_mode(&self) -> crate::ShutterMode {
-        self.active_controls().shutter_mode()
+        self.imp().camera_controls.shutter_mode()
     }
 
     pub fn set_shutter_mode(&self, shutter_mode: crate::ShutterMode) {
         if matches!(shutter_mode, crate::ShutterMode::Picture) {
             self.stop_recording();
         }
-        self.imp()
-            .camera_controls_horizontal
-            .set_shutter_mode(shutter_mode);
-        self.imp()
-            .camera_controls_vertical
-            .set_shutter_mode(shutter_mode);
+        self.imp().camera_controls.set_shutter_mode(shutter_mode);
+    }
+
+    fn set_detect_codes(&self, detect_codes: bool) {
+        let imp = self.imp();
+
+        imp.viewfinder.set_detect_codes(detect_codes);
+        imp.qr_screen_bin.set_enabled(detect_codes);
+
+        let layout = if detect_codes {
+            ControlsLayout::DetectingCodes
+        } else {
+            ControlsLayout::Default
+        };
+        imp.camera_controls.set_layout(layout);
     }
 
     pub fn set_gallery(&self, gallery: crate::Gallery) {
         let imp = self.imp();
 
-        imp.viewfinder.connect_picture_done(
-            glib::clone!(@weak gallery, @weak self as obj => move |_, file| {
+        imp.viewfinder.connect_picture_done(glib::clone!(
+            #[weak]
+            gallery,
+            #[weak(rename_to = obj)]
+            self,
+            move |_, file| {
                 let window = obj.root().and_downcast::<crate::Window>().unwrap();
                 window.set_shutter_enabled(true);
                 // TODO Maybe report error via toast on None
                 if let Some(file) = file {
                     gallery.add_image(file);
                 }
-            }),
-        );
-        imp.viewfinder.connect_recording_done(
-            glib::clone!(@weak gallery, @weak self as obj => move |_, file| {
-                let imp = obj.imp();
-                // TODO Maybe report error via toast on None
+            }
+        ));
+        imp.viewfinder.connect_recording_done(glib::clone!(
+            #[weak]
+            gallery,
+            move |_, file| {
                 if let Some(file) = file {
                     gallery.add_video(file);
+                } else {
+                    log::error!("Didn't find any file when recording finished!");
                 }
-                if matches!(imp.camera_controls_horizontal.shutter_mode(), crate::ShutterMode::Recording) {
-                    imp.camera_controls_horizontal.set_shutter_mode(crate::ShutterMode::Video);
-                    imp.camera_controls_vertical.set_shutter_mode(crate::ShutterMode::Video);
-                }
-            }),
-        );
-        imp.camera_controls_horizontal.set_gallery(&gallery);
-        imp.camera_controls_vertical.set_gallery(&gallery);
+            }
+        ));
+        imp.camera_controls.set_gallery(&gallery);
     }
 
     pub fn stop_stream(&self) {
@@ -393,12 +546,14 @@ impl Camera {
             .set_draw_guidelines(!imp.guidelines.draw_guidelines());
     }
 
+    pub fn is_recording_active(&self) -> bool {
+        self.imp().viewfinder.is_recording()
+    }
+
     fn update_cameras_button(&self, provider: &aperture::DeviceProvider) {
         let imp = self.imp();
 
-        imp.camera_controls_horizontal
-            .update_visible_camera_button(provider.n_items());
-        imp.camera_controls_vertical
+        imp.camera_controls
             .update_visible_camera_button(provider.n_items());
 
         // We need to set the correct selected item at least when loading. The
@@ -425,26 +580,19 @@ impl Camera {
         let imp = self.imp();
 
         if imp.permission_denied.get() {
-            imp.spinner.stop();
-            imp.stack.set_visible_child_name("not-found");
+            imp.stack.set_visible_child_name("permission-denied");
             return;
         }
 
         match imp.viewfinder.state() {
             aperture::ViewfinderState::Loading => {
-                imp.spinner.start();
                 imp.stack.set_visible_child_name("loading");
             }
             aperture::ViewfinderState::Ready => {
-                imp.spinner.stop();
                 imp.stack.set_visible_child_name("camera");
             }
-            aperture::ViewfinderState::NoCameras => {
-                imp.spinner.stop();
-                imp.stack.set_visible_child_name("not-found")
-            }
+            aperture::ViewfinderState::NoCameras => imp.stack.set_visible_child_name("not-found"),
             aperture::ViewfinderState::Error => {
-                imp.spinner.stop();
                 imp.stack.set_visible_child_name("camera");
 
                 let window = self.root().and_downcast::<crate::Window>().unwrap();
@@ -453,36 +601,31 @@ impl Camera {
         }
     }
 
-    fn update_window_controls(&self) {
-        let imp = self.imp();
-
-        let decoration_layout = self.settings().gtk_decoration_layout().and_then(|layout| {
-            layout
-                .split_once(':')
-                .map(|(_start, end)| end.split(',').rev().collect::<Vec<_>>().join(","))
-        });
-        imp.vertical_end_window_controls
-            .set_decoration_layout(decoration_layout.as_deref());
-    }
-
     fn show_recording_label(&self) {
         let imp = self.imp();
 
         let source = glib::timeout_add_seconds_local(
             1,
-            glib::clone!(@weak self as obj => @default-return glib::ControlFlow::Break, move || {
-                let imp = obj.imp();
+            glib::clone!(
+                #[weak(rename_to = obj)]
+                self,
+                #[upgrade_or]
+                glib::ControlFlow::Break,
+                move || {
+                    let imp = obj.imp();
 
-                // TODO Use Cell::update once stabilized.
-                let duration = imp.recording_duration.get() + 1;
-                imp.recording_duration.set(duration);
+                    // TODO Use Cell::update once stabilized.
+                    let duration = imp.recording_duration.get() + 1;
+                    imp.recording_duration.set(duration);
 
-                let minutes = duration.div_euclid(60);
-                let seconds = duration.rem_euclid(60);
-                imp.recording_label.set_label(&format!("{minutes}∶{seconds:02}"));
+                    let minutes = duration.div_euclid(60);
+                    let seconds = duration.rem_euclid(60);
+                    imp.recording_label
+                        .set_label(&format!("{minutes}∶{seconds:02}"));
 
-                glib::ControlFlow::Continue
-            }),
+                    glib::ControlFlow::Continue
+                }
+            ),
         );
         if let Some(old) = imp.recording_source.replace(Some(source)) {
             old.remove();
@@ -504,6 +647,25 @@ impl Camera {
     }
 }
 
+#[cfg(feature = "portal")]
+async fn stream() -> anyhow::Result<OwnedFd> {
+    let proxy = camera::Camera::new().await?;
+    proxy
+        .request_access()
+        .await
+        .context("org.freedesktop.portal.Camera.AccessCamera failed")?;
+    let is_present = proxy
+        .is_present()
+        .await
+        .context("org.freedesktop.portal.Camera.IsCameraPresent failed")?;
+    log::debug!("org.freedesktop.portal.Camera:IsCameraPresent: {is_present}");
+
+    proxy
+        .open_pipe_wire_remote()
+        .await
+        .context("org.freedesktop.portal.Camera.OpenPipeWireRemote")
+}
+
 // Id used to identify the last-used camera.
 fn id_from_pw(camera: &aperture::Camera) -> glib::GString {
     camera.display_name()
